@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2002-2005 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -20,9 +20,8 @@ import codecs
 import math
 import os
 import re
-import sys
-import tempfile
 import difflib
+import struct
 
 import pango
 import gobject
@@ -99,24 +98,18 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         """Start up an filediff with num_panes empty contents.
         """
         melddoc.MeldDoc.__init__(self, prefs)
-        gnomeglade.Component.__init__(self, paths.share_dir("glade2/filediff.glade"), "filediff")
+        override = {}
+        if sourceview_available:
+            override["GtkTextView"] = gsv.SourceView
+            override["GtkTextBuffer"] = gsv.SourceBuffer
+        gnomeglade.Component.__init__(self, paths.share_dir("glade2/filediff.glade"), "filediff", override)
         self._map_widgets_into_lists( ["textview", "fileentry", "diffmap", "scrolledwindow", "linkmap", "statusimage"] )
         self._update_regexes()
+        self.warned_bad_comparison = False
         if sourceview_available:
-            # ugly hack. http://bugzilla.gnome.org/show_bug.cgi?id=140071
-            # will remove the need for this.
-            self.textview = []
-            for w in self.scrolledwindow:
-                w.remove( w.get_child() )
-                v = gsv.SourceView()
-                self.textview.append( v )
-                v.show()
-                w.add(v)
+            for v in self.textview:
+                v.set_buffer( gsv.SourceBuffer() )
                 v.set_show_line_numbers(self.prefs.show_line_numbers)
-                for s in "key_press_event key_release_event".split():
-                    v.connect(s, getattr(self,"on_%s"%s))
-                for s in "button_press_event move_cursor focus_in_event".split():
-                    v.connect(s, getattr(self,"on_textview_%s"%s))
         self.keymask = 0
         self.load_font()
         self.deleted_lines_pending = -1
@@ -166,6 +159,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 self.parent.save()
             def on_save_as_activate(self, menuitem):
                 self.parent.save_file( self.pane, 1)
+            def on_make_patch_activate(self, menuitem):
+                self.parent.make_patch( self.pane )
             def on_cut_activate(self, menuitem):
                 self.parent.on_cut_activate()
             def on_copy_activate(self, menuitem):
@@ -189,8 +184,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         for r in [ misc.ListItem(i) for i in self.prefs.regexes.split("\n") ]:
             if r.active:
                 try:
-                    self.regexes.append( re.compile(r.value+"(?m)") )
-                except re.error, e:
+                    self.regexes.append( (re.compile(r.value+"(?m)"), r.value) )
+                except re.error:
                     pass
 
     def _disconnect_buffer_handlers(self):
@@ -214,15 +209,19 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     def _update_cursor_status(self, buf):
         def update():
-            iter = buf.get_iter_at_mark( buf.get_insert() )
-            status = "%s : %s" % ( _("INS,OVR").split(",")[ self.textview_overwrite ], #insert/overwrite
-                                   _("Ln %i, Col %i") % (iter.get_line()+1, iter.get_line_offset()+1) ) #line/column
+            it = buf.get_iter_at_mark( buf.get_insert() )
+            # Abbreviation for insert,overwrite so that it will fit in the status bar
+            insert_overwrite = _("INS,OVR").split(",")[ self.textview_overwrite ]
+            # Abbreviation for line, column so that it will fit in the status bar
+            line_column = _("Ln %i, Col %i") % (it.get_line()+1, it.get_line_offset()+1)
+            status = "%s : %s" % ( insert_overwrite, line_column )
             self.emit("status-changed", status  )
             raise StopIteration; yield 0
         self.scheduler.add_task( update().next )
 
-    def on_textview_move_cursor(self, view, *args):
-        self._update_cursor_status(view.get_buffer())
+    def on_textbuffer_mark_set(self, buffer, it, mark):
+        if mark.get_name() == "insert":
+            self._update_cursor_status(buffer)
     def on_textview_focus_in_event(self, view, event):
         self.textview_focussed = view
         self._update_cursor_status(view.get_buffer())
@@ -234,43 +233,56 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         if self.num_panes > 1:
             buffers = [t.get_buffer() for t in self.textview[:self.num_panes] ]
             pane = buffers.index(buffer)
-            range = self.linediffer.change_sequence( pane, startline, sizechange, self._get_texts())
-            for iter in self._update_highlighting( range[0], range[1] ):
+            change_range = self.linediffer.change_sequence( pane, startline, sizechange, self._get_texts())
+            for it in self._update_highlighting( change_range[0], change_range[1] ):
                 pass
             self.queue_draw()
         self._update_cursor_status(buffer)
 
     def _get_texts(self, raw=0):
-        class FakeTextRaw(object):
-            def __init__(self, buf, regexes):
-                self.buf = buf
-            def __getslice__(self, lo, hi):
-                b = self.buf
-                return b.get_text(b.get_iter_at_line(lo), b.get_iter_at_line(hi), 0).split("\n")[:-1]
-        class FakeTextFiltered(object):
-            def __init__(self, buf, regexes):
-                self.buf, self.regexes = buf, regexes
+        class FakeText(object):
+            def __init__(self, buf, textfilter):
+                self.buf, self.textfilter = buf, textfilter
             def __getslice__(self, lo, hi):
                 b = self.buf
                 txt = b.get_text(b.get_iter_at_line(lo), b.get_iter_at_line(hi), 0)
-                for r in self.regexes:
-                    txt = r.sub("", txt)
+                txt = self.textfilter(txt)
                 return txt.split("\n")[:-1]
-        FakeText = (FakeTextFiltered,FakeTextRaw)[raw]
         class FakeTextArray(object):
-            def __init__(self, bufs, regexes):
-                self.texts = [FakeText(b, regexes) for b in  bufs]
+            def __init__(self, bufs, textfilter):
+                self.texts = [FakeText(b, textfilter) for b in  bufs]
             def __getitem__(self, i):
                 return self.texts[i]
-        return FakeTextArray( [t.get_buffer() for t in self.textview], self.regexes )
+        return FakeTextArray( [t.get_buffer() for t in self.textview], [self._filter_text, lambda x:x][raw] )
 
-    def after_text_insert_text(self, buffer, iter, newtext, textlen):
+    def _filter_text(self, txt):
+        def killit(m):
+            assert m.group().count("\n") == 0
+            if len(m.groups()):
+                s = m.group()
+                for g in m.groups():
+                    if g:
+                        s = s.replace(g,"")
+                return s
+            else:
+                return ""
+        try:
+            for c,r in self.regexes:
+                txt = c.sub(killit,txt)
+        except AssertionError:
+            if not self.warned_bad_comparison:
+                misc.run_dialog(_("Regular expression '%s' changed the number of lines in the file. " \
+                    "Comparison will be incorrect. See the user manual for more details.") % r)
+                self.warned_bad_comparison = True
+        return txt
+
+    def after_text_insert_text(self, buffer, it, newtext, textlen):
         lines_added = newtext.count("\n")
-        starting_at = iter.get_line() - lines_added
+        starting_at = it.get_line() - lines_added
         self._after_text_modified(buffer, starting_at, lines_added)
 
-    def after_text_delete_range(self, buffer, iter0, iter1):
-        starting_at = iter0.get_line()
+    def after_text_delete_range(self, buffer, it0, it1):
+        starting_at = it0.get_line()
         assert self.deleted_lines_pending != -1
         self._after_text_modified(buffer, starting_at, -self.deleted_lines_pending)
         self.deleted_lines_pending = -1
@@ -391,17 +403,17 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
     def on_text_end_user_action(self, *buffer):
         self.undosequence.end_group()
 
-    def on_text_insert_text(self, buffer, iter, text, textlen):
+    def on_text_insert_text(self, buffer, it, text, textlen):
         if not self.undosequence_busy:
             self.undosequence.begin_group()
             pane = self.textview.index( buffer.textview )
             if self.bufferdata[pane].modified != 1:
                 self.undosequence.add_action( BufferModifiedAction(buffer, self) )
-            self.undosequence.add_action( BufferInsertionAction(buffer, iter.get_offset(), text) )
+            self.undosequence.add_action( BufferInsertionAction(buffer, it.get_offset(), text) )
             self.undosequence.end_group()
 
-    def on_text_delete_range(self, buffer, iter0, iter1):
-        text = buffer.get_text(iter0, iter1, 0)
+    def on_text_delete_range(self, buffer, it0, it1):
+        text = buffer.get_text(it0, it1, 0)
         pane = self.textview.index(buffer.textview)
         assert self.deleted_lines_pending == -1
         self.deleted_lines_pending = text.count("\n")
@@ -409,7 +421,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.undosequence.begin_group()
             if self.bufferdata[pane].modified != 1:
                 self.undosequence.add_action( BufferModifiedAction(buffer, self) )
-            self.undosequence.add_action( BufferDeletionAction(buffer, iter0.get_offset(), text) )
+            self.undosequence.add_action( BufferDeletionAction(buffer, it0.get_offset(), text) )
             self.undosequence.end_group()
 
         #
@@ -424,6 +436,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         return None
 
     def on_find_activate(self, *args):
+        self.keymask = 0
+        self.queue_draw()
         if self.find_dialog:
             self.find_dialog.widget.present()
         else:
@@ -475,6 +489,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             pane = self.textview.index(textview)
             self.popup_menu.popup_in_pane( pane )
             return 1
+        return 0
 
     def on_textview_toggle_overwrite(self, view):
         self.textview_overwrite = not self.textview_overwrite
@@ -508,10 +523,10 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             if match == None and wrap:
                 match = pattern.search(text, 0)
             if match:
-                iter = buf.get_iter_at_offset( match.start() )
-                buf.place_cursor( iter )
-                iter.forward_chars( match.end() - match.start() )
-                buf.move_mark( buf.get_selection_bound(), iter )
+                it = buf.get_iter_at_offset( match.start() )
+                buf.place_cursor( it )
+                it.forward_chars( match.end() - match.start() )
+                buf.move_mark( buf.get_selection_bound(), it )
                 view.scroll_to_mark( buf.get_insert(), 0)
             elif regex:
                 misc.run_dialog( _("The regular expression '%s' was not found.") % tofind_utf8, self, messagetype=gtk.MESSAGE_INFO)
@@ -544,7 +559,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         """Set num panes to len(files) and load each file given.
            If an element is None, the text of a pane is left as is.
         """
-        for i,f in misc.enumerate(files):
+        for i,f in enumerate(files):
             if f:
                 b = self.textview[i].get_buffer()
                 b.delete( b.get_start_iter(), b.get_end_iter() )
@@ -565,7 +580,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         yield _("[%s] Opening files") % self.label_text
         panetext = ["\n"] * self.num_panes
         tasks = []
-        for i,f in misc.enumerate(files):
+        for i,f in enumerate(files):
             if f:
                 try:
                     task = misc.struct(filename = f,
@@ -578,7 +593,10 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 except (IOError, LookupError), e:
                     buffers[i].set_text("\n")
                     misc.run_dialog(
-                        _("Could not open '%s' for reading.\n\nThe error was:\n%s") % (f, str(e)),
+                        "%s\n\n%s\n%s" % (
+                            _("Could not read from '%s'") % f,
+                            _("The error was:"),
+                            str(e)),
                         parent = self)
             else:
                 panetext[i] = buffers[i].get_text( buffers[i].get_start_iter(), buffers[i].get_end_iter() )
@@ -587,6 +605,14 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             for t in tasks[:]:
                 try:
                     nextbit = t.file.read(4096)
+                    if nextbit.find("\x00") != -1:
+                        misc.run_dialog(
+                            "%s\n\n%s" % (
+                                _("Could not read from '%s'") % t.filename,
+                                _("It contains ascii nulls.\nPerhaps it is a binary file.") ),
+                                parent = self )
+                        t.buf.delete( t.buf.get_start_iter(), t.buf.get_end_iter() )
+                        tasks.remove(t)
                 except ValueError, err:
                     t.codec.pop(0)
                     if len(t.codec):
@@ -597,12 +623,17 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                         print "codec error fallback", err
                         t.buf.delete( t.buf.get_start_iter(), t.buf.get_end_iter() )
                         misc.run_dialog(
-                            _("Could not read from '%s'.\n\nI tried encodings %s.") % (t.filename, try_codecs),
+                            "%s\n\n%s" % (
+                                _("Could not read from '%s'") % t.filename,
+                                _("I tried encodings %s.") % try_codecs ),
                             parent = self)
                         tasks.remove(t)
                 except IOError, ioerr:
                     misc.run_dialog(
-                        _("Could not read from '%s'.\n\nThe error was:\n%s") % (t.filename, str(ioerr)),
+                        "%s\n\n%s\n%s" % (
+                            _("Could not read from '%s'") % t.filename,
+                            _("The error was:"),
+                            str(ioerr)),
                         parent = self)
                     tasks.remove(t)
                 else:
@@ -623,8 +654,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             yield 1
         self.undosequence.clear()
         yield _("[%s] Computing differences") % self.label_text
-        for r in self.regexes:
-            panetext = [r.sub("",p) for p in panetext]
+        panetext = [self._filter_text(p) for p in panetext]
         lines = map(lambda x: x.split("\n"), panetext)
         step = self.linediffer.set_sequences_iter(*lines)
         while step.next() == None:
@@ -648,74 +678,113 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 tag = table.lookup(tagname)
                 b.remove_tag(tag, b.get_start_iter(), b.get_end_iter() )
         for chunk in self.linediffer.all_changes(self._get_texts()):
-            for i,c in misc.enumerate(chunk):
-                if c:
-                    if c[0] == "insert":
-                        buf = buffers[i*2]
-                        #txt = self._get_texts()[0][c[3]:c[4]]
-                        #print txt, "".join(txt) == ""
-                        #if "".join(txt) == "": continue
-                        #print "OK"
-                        tag = buf.get_tag_table().lookup("delete line")
-                        buf.apply_tag( tag, buf.get_iter_at_line(c[3]), buf.get_iter_at_line(c[4]) )
-                    elif c[0] == "delete":
-                        buf = buffers[1]
-                        tag = buf.get_tag_table().lookup("delete line")
-                        buf.apply_tag( tag, buf.get_iter_at_line(c[1]), buf.get_iter_at_line(c[2]) )
-                    elif c[0] == "conflict":
-                        bufs = buffers[1], buffers[i*2]
-                        tags = [b.get_tag_table().lookup("conflict line") for b in bufs]
-                        for b,t,o in zip(bufs, tags, (0,2)):
-                            b.apply_tag( t, b.get_iter_at_line(c[o+1]), b.get_iter_at_line(c[o+2]) )
-                    elif c[0] == "replace":
-                        bufs = buffers[1], buffers[i*2]
-                        tags = [b.get_tag_table().lookup("replace line") for b in bufs]
-                        starts = [b.get_iter_at_line(l) for b,l in zip(bufs, (c[1],c[3])) ]
-                        for b, t, s, l in zip(bufs, tags, starts, (c[2],c[4])):
-                            b.apply_tag(t, s, b.get_iter_at_line(l))
-                        if 1:
-                            text1 = "\n".join( self._get_texts(raw=1)[1  ][c[1]:c[2]] )
-                            textn = "\n".join( self._get_texts(raw=1)[i*2][c[3]:c[4]] )
-                            matcher = difflib.SequenceMatcher(None, text1, textn)
-                            #print "<<<\n%s\n---\n%s\n>>>" % (text1, textn)
-                            tags = [b.get_tag_table().lookup("inline line") for b in bufs]
-                            back = (0,0)
-                            for o in matcher.get_opcodes():
-                                if o[0] == "equal":
-                                    if (o[2]-o[1] < 3) or (o[4]-o[3] < 3):
-                                        back = o[4]-o[3], o[2]-o[1]
-                                    continue
-                                for i in range(2):
-                                    s,e = starts[i].copy(), starts[i].copy()
-                                    s.forward_chars( o[1+2*i] - back[i] )
-                                    e.forward_chars( o[2+2*i] )
-                                    bufs[i].apply_tag(tags[i], s, e)
-                                back = (0,0)
-                            yield 1
-        yield 1
+            for i,c in enumerate(chunk):
+                if c and c[0] == "replace":
+                    bufs = buffers[1], buffers[i*2]
+                    #tags = [b.get_tag_table().lookup("replace line") for b in bufs]
+                    starts = [b.get_iter_at_line(l) for b,l in zip(bufs, (c[1],c[3])) ]
+                    text1 = "\n".join( self._get_texts(raw=1)[1  ][c[1]:c[2]] ).encode("utf16")
+                    text1 = struct.unpack("%iH"%(len(text1)/2), text1)[1:]
+                    textn = "\n".join( self._get_texts(raw=1)[i*2][c[3]:c[4]] ).encode("utf16")
+                    textn = struct.unpack("%iH"%(len(textn)/2), textn)[1:]
+                    matcher = difflib.SequenceMatcher(None, text1, textn)
+                    #print "<<<\n%s\n---\n%s\n>>>" % (text1, textn)
+                    tags = [b.get_tag_table().lookup("inline line") for b in bufs]
+                    back = (0,0)
+                    for o in matcher.get_opcodes():
+                        if o[0] == "equal":
+                            if (o[2]-o[1] < 3) or (o[4]-o[3] < 3):
+                                back = o[4]-o[3], o[2]-o[1]
+                            continue
+                        for i in range(2):
+                            s,e = starts[i].copy(), starts[i].copy()
+                            s.forward_chars( o[1+2*i] - back[i] )
+                            e.forward_chars( o[2+2*i] )
+                            bufs[i].apply_tag(tags[i], s, e)
+                        back = (0,0)
+                    yield 1
+ 
+    def on_textview_expose_event(self, textview, event):
+        if self.num_panes == 1:
+            return
+        if event.window != textview.get_window(gtk.TEXT_WINDOW_TEXT) \
+            and event.window != textview.get_window(gtk.TEXT_WINDOW_LEFT):
+            return
+        if not hasattr(textview, "meldgc"):
+            self._setup_gcs(textview)
+        visible = textview.get_visible_rect()
+        pane = self.textview.index(textview)
+        start_line = self._pixel_to_line(pane, visible.y)
+        end_line = 1+self._pixel_to_line(pane, visible.y+visible.height)
+        gc = lambda x : getattr(textview.meldgc, "gc_"+x)
+        #gcdark = textview.get_style().black_gc
+        gclight = textview.get_style().bg_gc[gtk.STATE_ACTIVE]
+        #curline = textview.get_buffer().get_iter_at_mark( textview.get_buffer().get_insert() ).get_line()
+               
+        def draw_change(change): # draw background and thin lines
+            ypos0 = self._line_to_pixel(pane, change[1]) - visible.y
+            width = event.window.get_size()[0]
+            #gcline = (gclight, gcdark)[change[1] <= curline and curline < change[2]]
+            gcline = gclight
+            event.window.draw_line(gcline, 0,ypos0-1, width,ypos0-1)
+            if change[2] != change[1]:
+                ypos1 = self._line_to_pixel(pane, change[2]) - visible.y
+                event.window.draw_line(gcline, 0,ypos1, width,ypos1)
+                event.window.draw_rectangle(gc(change[0]), 1, 0,ypos0, width,ypos1-ypos0)
+        last_change = None
+        for change in self.linediffer.single_changes(pane, self._get_texts()):
+            if change[2] < start_line: continue
+            if change[1] > end_line: break
+            if last_change and change[1] <= last_change[2]:
+                last_change = ("conflict", last_change[1], max(last_change[2],change[2]))
+            else:
+                if last_change:
+                    draw_change(last_change)
+                last_change = change
+        if last_change:
+            draw_change(last_change)
         
+    def _get_filename_for_saving(self, title ):
+        dialog = gtk.FileChooserDialog(title,
+            parent=self.widget.get_toplevel(),
+            action=gtk.FILE_CHOOSER_ACTION_SAVE,
+            buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK) )
+        response = dialog.run()
+        filename = None
+        if response == gtk.RESPONSE_OK:
+            filename = dialog.get_filename()
+        dialog.destroy()
+        if filename:
+            if os.path.exists(filename):
+                response = misc.run_dialog(
+                    _('"%s" exists!\nOverwrite?') % os.path.basename(filename),
+                    parent = self,
+                    buttonstype = gtk.BUTTONS_YES_NO)
+                if response == gtk.RESPONSE_NO:
+                    return None
+            return filename
+        return None
+
+    def _save_text_to_filename(self, filename, text):
+        try:
+            open(filename, "w").write(text)
+        except IOError, e:
+            misc.run_dialog(
+                _("Error writing to %s\n\n%s.") % (filename, e),
+                self, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK)
+            return False
+        return True
+
     def save_file(self, pane, saveas=0):
         buf = self.textview[pane].get_buffer()
         bufdata = self.bufferdata[pane]
         if saveas or not bufdata.filename:
-            fselect = gtk.FileSelection( _("Choose a name for buffer %i.") % (pane+1))
-            fselect.set_transient_for(self.widget.get_toplevel() )
-            response = fselect.run()
-            if response != gtk.RESPONSE_OK:
-                fselect.destroy()
-                return melddoc.RESULT_ERROR
-            else:
-                filename = fselect.get_filename()
-                fselect.destroy()
-                if os.path.exists(filename):
-                    response = misc.run_dialog(
-                        _('"%s" exists!\nOverwrite?') % os.path.basename(filename),
-                        parent = self,
-                        buttonstype = gtk.BUTTONS_YES_NO)
-                    if response == gtk.RESPONSE_NO:
-                        return melddoc.RESULT_ERROR
+            filename = self._get_filename_for_saving( _("Choose a name for buffer %i.") % (pane+1) )
+            if filename:
                 bufdata.filename = os.path.abspath(filename)
                 self.fileentry[pane].set_filename( bufdata.filename)
+            else:
+                return melddoc.RESULT_ERROR
         text = buf.get_text(buf.get_start_iter(), buf.get_end_iter(), 0)
         if bufdata.newlines:
             if type(bufdata.newlines) == type(""):
@@ -737,23 +806,46 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         if bufdata.encoding and self.prefs.save_encoding==0:
             try:
                 text = text.encode(bufdata.encoding)
-            except UnicodeEncodeError, e:
+            except UnicodeEncodeError:
                 if misc.run_dialog(
                     _("'%s' contains characters not encodable with '%s'\nWould you like to save as UTF-8?") % (bufdata.filename, bufdata.encoding),
                     self, gtk.MESSAGE_ERROR, gtk.BUTTONS_YES_NO) != gtk.RESPONSE_YES:
                     return melddoc.RESULT_ERROR
-        try:
-            open(bufdata.filename, "w").write(text)
-        except IOError, e:
-            misc.run_dialog(
-                _("Error writing to %s\n\n%s.") % (bufdata.filename, e),
-                self, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK)
-            return melddoc.RESULT_ERROR
-        else:
+        if self._save_text_to_filename(bufdata.filename, text):
             self.emit("file-changed", bufdata.filename)
             self.undosequence.clear()
             self.set_buffer_modified(buf, 0)
-        return melddoc.RESULT_OK
+            return melddoc.RESULT_OK
+        else:
+            return melddoc.RESULT_ERROR
+
+    def make_patch(self, pane):
+        fontdesc = pango.FontDescription(self.prefs.get_current_font())
+        dialog = gnomeglade.Component( paths.share_dir("glade2/filediff.glade"), "patchdialog")
+        dialog.widget.set_transient_for( self.widget.get_toplevel() )
+        bufs = [t.get_buffer() for t in self.textview]
+        texts = [b.get_text(*b.get_bounds()).split("\n") for b in bufs]
+        texts[0] = [l+"\n" for l in texts[0]]
+        texts[1] = [l+"\n" for l in texts[1]]
+        names = [self._get_filename(i) for i in range(2)]
+        dialog.textview.modify_font(fontdesc)
+        buf = dialog.textview.get_buffer()
+        lines = []
+        for line in difflib.unified_diff(texts[0], texts[1], names[0], names[1]):
+            buf.insert( buf.get_end_iter(), line )
+            lines.append(line)
+        result = dialog.widget.run()
+        dialog.widget.destroy()
+        if result >= 0:
+            txt = "".join(lines)
+            if result == 1: # copy
+                clip = gtk.clipboard_get()
+                clip.set_text(txt)
+                clip.store()
+            else:# save as
+                filename = self._get_filename_for_saving( _("Save patch as...") )
+                if filename:
+                    self._save_text_to_filename(filename, txt)
 
     def set_buffer_writable(self, buf, yesno):
         pane = self.textview.index(buf.textview)
@@ -827,7 +919,6 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         if not self._sync_hscroll_lock:
             self._sync_hscroll_lock = 1
             adjs = map( lambda x: x.get_hadjustment(), self.scrolledwindow)
-            master = adjs.index(adjustment)
             adjs.remove(adjustment)
             val = adjustment.get_value()
             for a in adjs:
@@ -902,7 +993,6 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         scaleit = lambda x,s=hperline,o=size_of_arrow: x*s+o
         x0 = 4
         x1 = area.get_allocation().width - 2*x0
-        madj = self.scrolledwindow[textindex].get_vadjustment()
 
         window = area.window
         window.clear()
@@ -930,6 +1020,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             upper = adj.upper - adj.page_size
             adj.set_value( max( min(upper, val), 0) )
             return 1
+        return 0
 
     def _get_line_count(self, index):
         """Return the number of lines in the buffer of textview 'text'"""
@@ -953,8 +1044,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.recompute_label()
 
     def _line_to_pixel(self, pane, line ):
-        iter = self.textview[pane].get_buffer().get_iter_at_line(line)
-        return self.textview[pane].get_iter_location( iter ).y
+        it = self.textview[pane].get_buffer().get_iter_at_line(line)
+        return self.textview[pane].get_iter_location( it ).y
 
     def _pixel_to_line(self, pane, pixel ):
         return self.textview[pane].get_line_at_y( pixel )[0].get_line()
@@ -1000,6 +1091,20 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         area.meldgc = misc.struct(gc_delete=gcd, gc_insert=gcd, gc_replace=gcc, gc_conflict=gcx)
         area.meldgc.get_gc = lambda p: getattr(area.meldgc, "gc_"+p)
 
+    def _consume_blank_lines(self, txt):
+        lo, hi = 0, 0
+        for l in txt:
+            if len(l)==0:
+                lo += 1
+            else:
+                break
+        for l in txt[lo:]:
+            if len(l)==0:
+                hi += 1
+            else:
+                break
+        return lo,hi
+
         #
         # linkmap drawing
         #
@@ -1009,7 +1114,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         if not window: return
         if not hasattr(area, "meldgc"):
             self._setup_gcs(area)
-        gctext = area.get_style().text_gc[0]
+        gctext = area.get_style().bg_gc[gtk.STATE_ACTIVE]
 
         alloc = area.get_allocation()
         (wtotal,htotal) = alloc.width, alloc.height
@@ -1047,24 +1152,10 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             return [self._pixel_to_line(idx, pix_start[idx]), self._pixel_to_line(idx, pix_start[idx]+htotal)]
         visible = [None] + bounds(which) + bounds(which+1)
 
-        def consume_blank_lines(txt):
-            lo, hi = 0, 0
-            for l in txt:
-                if len(l)==0:
-                    lo += 1
-                else:
-                    break
-            for l in txt[lo:]:
-                if len(l)==0:
-                    hi += 1
-                else:
-                    break
-            return lo,hi
-
         for c in self.linediffer.pair_changes(which, which+1, self._get_texts()):
             if self.prefs.ignore_blank_lines:
-                c1,c2 = consume_blank_lines( self._get_texts()[which  ][c[1]:c[2]] )
-                c3,c4 = consume_blank_lines( self._get_texts()[which+1][c[3]:c[4]] )
+                c1,c2 = self._consume_blank_lines( self._get_texts()[which  ][c[1]:c[2]] )
+                c3,c4 = self._consume_blank_lines( self._get_texts()[which+1][c[3]:c[4]] )
                 c = c[0], c[1]+c1,c[2]-c2, c[3]+c3,c[4]-c4
                 if c[1]==c[2] and c[3]==c[4]:
                     continue
@@ -1103,12 +1194,12 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
 
             x = wtotal-self.pixbuf_apply0.get_width()
             if c[0]=="insert":
-                pix1.render_to_drawable( window, gctext, 0,0, x, points0[-1][1], -1,-1, 0,0,0)
+                window.draw_pixbuf( gctext, pix1, 0,0, x, points0[-1][1], -1,-1, 0,0,0)
             elif c[0] == "delete":
-                pix0.render_to_drawable( window, gctext, 0,0, 0, points0[ 0][1], -1,-1, 0,0,0)
+                window.draw_pixbuf( gctext, pix0, 0,0, 0, points0[ 0][1], -1,-1, 0,0,0)
             else: #replace
-                pix0.render_to_drawable( window, gctext, 0,0, 0, points0[ 0][1], -1,-1, 0,0,0)
-                pix1.render_to_drawable( window, gctext, 0,0, x, points0[-1][1], -1,-1, 0,0,0)
+                window.draw_pixbuf( gctext, pix0, 0,0, 0, points0[ 0][1], -1,-1, 0,0,0)
+                window.draw_pixbuf( gctext, pix1, 0,0, x, points0[-1][1], -1,-1, 0,0,0)
 
         # allow for scrollbar at end of textview
         mid = 0.5 * self.textview0.get_allocation().height
@@ -1151,6 +1242,12 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             func = lambda c: self._line_to_pixel(src, c[1]) - adj.value
 
             for c in self.linediffer.pair_changes(src, dst, self._get_texts()):
+                if self.prefs.ignore_blank_lines:
+                    c1,c2 = self._consume_blank_lines( self._get_texts()[src][c[1]:c[2]] )
+                    c3,c4 = self._consume_blank_lines( self._get_texts()[dst][c[3]:c[4]] )
+                    c = c[0], c[1]+c1,c[2]-c2, c[3]+c3,c[4]-c4
+                    if c[1]==c[2] and c[3]==c[4]:
+                        continue
                 if c[0] == "insert":
                     continue
                 h = func(c)
@@ -1165,6 +1262,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             return 1
         elif event.button == 2:
             self.linkmap_drag_coord = event.x
+        return 0
 
     def on_linkmap_motion_notify_event(self, area, event):
         return 
@@ -1213,6 +1311,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                         b1.insert_with_tags_by_name(b1.get_iter_at_line(chunk[2]), t0, "edited line")
                         self.on_text_end_user_action()
             return 1
+        return 0
 
     def on_linkmap_drag_begin(self, *args):
         print args
